import json
import uuid
from typing import Generic, Optional, TypeVar, Union

from attr import asdict, define

from qtgql.utils.graphql import get_operation_name
from qtgql.utils.typingref import UNSET


@define
class EncodeAble:
    def asdict(self) -> dict:
        return {k: v for k, v in asdict(self).items() if v is not UNSET}


T = TypeVar("T")


@define(kw_only=True)
class QueryPayload(Generic[T], EncodeAble):
    query: str
    operationName: Optional[str] = UNSET
    variables: Optional[Union[dict, T]] = UNSET

    def __attrs_post_init__(self):
        if not self.operationName:
            self.operationName = get_operation_name(self.query)


@define(kw_only=True, repr=False)
class SafeGqlPayload(QueryPayload):
    ...


@define(kw_only=True)
class GqlResult:
    data: dict
    errors: Optional[list[dict]] = None


class GqlEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):  # noqa: R505
            return str(obj)
        elif isinstance(obj, (EncodeAble, QueryPayload)):
            return obj.asdict()
        return None



from typing import Any, Type


class AntiForwardRef:
    """
    i.e:
    Union["someString"] would return a ForwardRef, this class is a simple hack
    to just return a type contains the name.
    """

    name: str
    type_map: dict

    @classmethod
    def resolve(cls) -> Any:
        return cls.type_map[cls.name]


def anti_forward_ref(name: str, type_map: dict) -> Type[AntiForwardRef]:
    return type(name, (AntiForwardRef,), {"name": name, "type_map": type_map})



from pathlib import Path
from typing import Callable, Type

from attrs import define

from qtgql.codegen.introspection import SchemaEvaluator
from qtgql.codegen.py.compiler.template import TemplateContext, schema_types_template
from qtgql.codegen.py.runtime.bases import BaseGraphQLObject, _BaseQGraphQLObject
from qtgql.codegen.py.runtime.custom_scalars import CUSTOM_SCALARS, CustomScalarMap


@define
class QtGqlConfig:
    """Encapsulates configurations for a qtgql-codegen application per GraphQL
    schema."""

    graphql_dir: Path
    """A directory contains [schema.graphql, query.graphql, mutation.graphql,
    subscription.graphql] generated types come from the schema.

    and queries, mutations, subscription handlers would be generated
    from the corresponding `.graphql` files.
    """
    env_name: str = "QGqlEnv"
    """The generated types would find the environment by this name.

    Also the generated QML imports would fall under this namespace.
    """
    evaluator: Type[SchemaEvaluator] = SchemaEvaluator
    """evaluates the schema and generates types."""
    custom_scalars: CustomScalarMap = CUSTOM_SCALARS
    """mapping of custom scalars, respected by the schema evaluator."""
    template_class: Callable[[TemplateContext], str] = schema_types_template
    """jinja template."""
    base_object: Type[_BaseQGraphQLObject] = BaseGraphQLObject
    """base object to be extended by all generated types."""

    @property
    def schema_path(self) -> Path:
        return self.graphql_dir / "schema.graphql"

    @property
    def operations_dir(self) -> Path:
        return self.graphql_dir / "operations.graphql"

    @property
    def generated_types_dir(self) -> Path:
        return self.graphql_dir / "objecttypes.py"

    @property
    def generated_handlers_dir(self) -> Path:
        return self.graphql_dir / "handlers.py"

    def generate(self) -> None:
        self.evaluator(self).dump()

    def __attrs_post_init__(self):
        if self.custom_scalars != CUSTOM_SCALARS:
            self.custom_scalars.update(CUSTOM_SCALARS)



import types
from typing import Any, Type, TypeVar, Union, cast, get_args

import attr
from attr import Attribute, define
from PySide6 import QtCore as qtc
from typing_extensions import dataclass_transform

__all__ = ["define_properties"]


def signal_name(name: str) -> str:
    return name + "Changed"


class PropertyMeta(type(qtc.QObject)):  # type: ignore
    def __new__(cls, name, bases, attrs):
        return super().__new__(cls, name, bases, attrs)


class BaseInner(qtc.QObject):
    def __init__(self, attrs_instance, parent=None):
        super().__init__(parent)
        self.attrs_instance = attrs_instance


class PropertyImpl(qtc.Property):
    def __init__(self, type_: type, name: str, notify: qtc.Signal):
        self.name = name
        super().__init__(type_, fget=self.getter, fset=self.setter, notify=notify)  # type: ignore

    def getter(self, instance: BaseInner):  # type: ignore
        return getattr(instance.attrs_instance, f"{self.name}")

    def setter(self, instance: BaseInner, value):  # type: ignore
        signal = getattr(instance, signal_name(self.name))
        setattr(instance.attrs_instance, f"{self.name}", value)
        signal.emit()


class MakeProperties:
    PropertiesClass_: Type[BaseInner]
    properties: BaseInner

    def __init_subclass__(cls, **kwargs):
        annotations: dict[str, type] = cls.__annotations__
        properties: dict[str, Union[PropertyImpl, qtc.Signal]] = {}
        signals: dict[str, qtc.Signal] = {}
        for key in cls.__dict__.keys():
            if key in annotations:
                type_ = annotations[key]
                if args := get_args(type_):
                    type_ = args[0]
                if type_ is Any:
                    type_ = "QVariant"
                signals[signal_name(key)] = signal = qtc.Signal()
                properties[key] = PropertyImpl(type_=type_, name=key, notify=signal)

        inner = types.new_class(
            name=cls.__name__ + "Properties",
            bases=(BaseInner,),
            kwds={"metaclass": PropertyMeta},
            exec_body=lambda ns: ns.update(dict(**signals, **properties)),
        )
        cls.PropertiesClass_ = inner


def attrs_compat_setter(instance: "BasePropertyClass", attr: Attribute, value):
    getattr(instance.properties, signal_name(attr.name)).emit()
    return value


T = TypeVar("T")

BasePropertyClass = TypeVar("BasePropertyClass", bound=MakeProperties)


@dataclass_transform(
    field_descriptors=(attr.attrib, attr.field),
)
def define_properties(cls: Union[Type[T], Type[MakeProperties]]) -> Type[BasePropertyClass]:
    """
    :param cls: Class that implements dataclass syntax - (PEP 681).
    :return: `@attrs.define` decorated class with properties field injected.
     this field is a `QObject` that will emit signals when you set attribute on that class
     and is use-able via QML and other parts of QT.
    """
    ns = cls.__dict__.copy()
    bases = cls.__bases__ if cls.__bases__ != (object,) else (MakeProperties,)
    cls = type(cls.__name__, bases, ns)

    def post_init(self: BasePropertyClass):
        self.properties = self.PropertiesClass_(self)

    cls.__attrs_post_init__ = post_init  # type: ignore
    return cast(Type[BasePropertyClass], define(cls, on_setattr=attrs_compat_setter))



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from qtgql.codegen.py.objecttype import GqlFieldDefinition

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.introspection import OperationName


class QueryHandlerDefinition(NamedTuple):
    query: str
    name: OperationName
    field: GqlFieldDefinition
    directives: list[str] = []
    fragments: list[str] = []



import inspect
from typing import Any, List, Optional, Type, TypeVar, Union, get_args, get_origin


class UnsetType:
    __instance: Optional["UnsetType"] = None

    def __new__(cls: Type["UnsetType"]) -> "UnsetType":
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            return cls.__instance
        return cls.__instance

    def __str__(self):
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return "UNSET"

    def __bool__(self):
        return False


UNSET: Any = UnsetType()


class TypeHinter:
    def __init__(
        self,
        type: Any,  # noqa: A003
        of_type: tuple["TypeHinter", ...] = (),
    ):
        self.type = type
        self.of_type = of_type

    def __eq__(self, other) -> bool:
        if not isinstance(other, TypeHinter):
            return False

        if not self.type == other.type:
            return False

        for count, child in enumerate(self.of_type):
            if not child == other.of_type[count]:
                return False

        return True

    @classmethod
    def from_string(cls, tp: str, ns: dict) -> "TypeHinter":
        from typing import List, Union, Dict, Optional, Any, Type, get_type_hints  # noqa

        ns = ns.copy()
        ns.update(locals())

        def tmp(t: tp):  # type: ignore
            ...  # pragma: no cover

        annotation = get_type_hints(tmp, localns=ns, globalns=globals())["t"]

        return cls.from_annotations(annotation)

    @classmethod
    def from_annotations(cls, tp: Any) -> "TypeHinter":
        if args := get_args(tp):
            # handle optional
            if type(None) in args:
                # optional union, 2 is default for optionals
                if len(args) > 2:
                    return cls(
                        type=Optional,
                        of_type=(
                            cls(
                                type=Union,
                                of_type=tuple(
                                    cls.from_annotations(arg)
                                    for arg in args
                                    if arg is not type(None)
                                ),
                            ),
                        ),
                    )
                return cls(type=Optional, of_type=(cls.from_annotations(args[0]),))
            new_args: list[TypeHinter] = []
            for arg in args:
                new_args.append(cls.from_annotations(arg))

            return cls(type=get_origin(tp), of_type=tuple(new_args))  # type: ignore
        return cls(type=tp)

    def as_annotation(self, object_map: Optional[dict[str, Any]] = None) -> Any:
        if self.type is str:
            return self.type
        # eval forward refs
        if isinstance(self.type, str):
            assert object_map, "can't evaluate forward refs without object_map."
            self.type = object_map[self.type]

        if builder := getattr(
            self.type, "__class_getitem__", getattr(self.type, "__getitem__", None)
        ):
            if self.is_union():
                return builder(tuple(arg.as_annotation(object_map) for arg in self.of_type))
            return builder(self.of_type[0].as_annotation(object_map))
        return self.type

    def stringify(self) -> str:
        annot = self.as_annotation()
        if inspect.isclass(annot) and not hasattr(annot, "__origin__"):
            return annot.__name__

        return str(annot).replace("typing.", "").replace("qtgql.codegen.utils.", "")

    def is_union(self) -> bool:
        return self.type is Union

    def is_optional(self) -> bool:
        return self.type is Optional

    def is_list(self) -> bool:
        return self.type in (list, List)

    @classmethod
    def strip_optionals(cls, inst: "TypeHinter") -> "TypeHinter":
        th = inst
        if inst.is_optional():
            th = inst.of_type[0]

        return cls(type=th.type, of_type=tuple(cls.strip_optionals(tp) for tp in th.of_type))


T = TypeVar("T")


def ensure(inst: T, tp: Type[T]) -> T:
    if not isinstance(inst, tp):
        raise TypeError(f"{inst} is not of type {tp}")
    return inst



from abc import ABC, abstractmethod
from datetime import date, datetime, time
from typing import Any, Generic, Optional, Type, TypeVar

from _decimal import Decimal

T = TypeVar("T")

__all__ = ["BaseCustomScalar"]


class BaseCustomScalar(Generic[T], ABC):
    """Class to extend by user defined scalars."""

    __slots__ = "_value"
    GRAPHQL_NAME: str
    """The *real* GraphQL name of the scalar (used by the codegen inspection
    pipeline)."""
    DEFAULT_VALUE: T
    """A place holder graphql query returned null or the field wasn't queried.

    can be used by `from_graphql()`
    """

    def __init__(self, v: Optional[T] = None):
        if not v:
            self._value = self.DEFAULT_VALUE
        else:
            self._value = v

    @abstractmethod
    def from_graphql(cls, v: Optional[Any] = None) -> "BaseCustomScalar":
        """Deserializes data fetched from graphql, This is useful when you want
        to set a first-of value that will later be used by `to_qt()`.

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_qt(self) -> Any:
        """Will be used by the property getter, This is the official value that
        Qt should "understand".

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover


class DateTimeScalar(BaseCustomScalar[datetime]):
    """An ISO-8601 encoded datetime."""

    GRAPHQL_NAME: str = "DateTime"
    DEFAULT_VALUE = datetime.now()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateTimeScalar":
        if v:
            return cls(datetime.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.strftime("%H:%M (%m/%d/%Y)")


class DateScalar(BaseCustomScalar[date]):
    """An ISO-8601 encoded date."""

    GRAPHQL_NAME = "Date"
    DEFAULT_VALUE = date(year=1998, month=8, day=23)

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateScalar":
        if v:
            return cls(date.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class TimeScalar(BaseCustomScalar[time]):
    """an ISO-8601 encoded time."""

    GRAPHQL_NAME = "Time"
    DEFAULT_VALUE = time()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "TimeScalar":
        if v:
            return cls(time.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class DecimalScalar(BaseCustomScalar[Decimal]):
    """A Decimal value serialized as a string."""

    GRAPHQL_NAME = "Decimal"
    DEFAULT_VALUE = Decimal()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DecimalScalar":
        if v:
            return cls(Decimal(v))
        return cls()

    def to_qt(self) -> str:
        return str(self._value)


CustomScalarMap = dict[str, Type[BaseCustomScalar]]
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalar.GRAPHQL_NAME: DateTimeScalar,
    DecimalScalar.GRAPHQL_NAME: DecimalScalar,
    DateScalar.GRAPHQL_NAME: DateScalar,
    TimeScalar.GRAPHQL_NAME: TimeScalar,
}



from __future__ import annotations
from PySide6.QtCore import Signal, QObject, QEnum

from PySide6.QtQuick import QQuickItem
from typing import Optional, Union
from enum import Enum, auto
from PySide6.QtQml import QmlElement, QmlSingleton

from qtgql.tools import qproperty
from qtgql.codegen.py.runtime.bases import QGraphQListModel






{% for dep in context.dependencies %}
{{dep}}{% endfor %}


QML_IMPORT_NAME = "generated.{{context.config.env_name}}.types"
QML_IMPORT_MAJOR_VERSION = 1



{% for enum in context.enums %}
class {{enum.name}}(Enum):
    {% for member in enum.members %}
    {{member.name}} = auto()
    """{{member.description}}"""{% endfor %}

{% endfor %}

{% if context.enums %}
@QmlElement
class Enums(QObject):
    {% for enum in context.enums %}
    QEnum({{enum.name}})
    {% endfor %}
{% endif %}


class SCALARS:
    {% for scalar in context.custom_scalars %}
    {{scalar}} = {{scalar}}{% endfor %}

{% for type in context.types %}
class {{ type.name }}({{context.base_object_name}}):
    """{{  type.docstring  }}"""

    DEFAULT_INIT_DICT = dict({% for f in type.fields %}
        {{f.name}}=None,{% endfor %}
    )
    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: {{f.annotation}} = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{  f.private_name  }} = {{f.name}} if {{f.name}} else {{f.default_value}}{% endfor %}

    @classmethod
    def from_dict(cls, parent,  data: dict) -> {{type.name}}:
        init_dict = cls.DEFAULT_INIT_DICT.copy()
        {% for f in type.fields %}
        if {{f.name}} := data.get('{{f.name}}', None):
            init_dict['{{f.name}}'] = {{f.deserializer}}{% endfor %}
        return cls(
            parent=parent,
            **init_dict
        )

    {% for f in type.fields %}
    {{ f.signal_name }} = Signal()

    def {{ f.setter_name }}(self, v: {{  f.annotation  }}) -> None:
        self.{{  f.private_name  }} = v
        self.{{  f.name  }}Changed.emit()

    @qproperty(type={{f.property_type}}, fset={{ f.setter_name }}, notify={{f.signal_name}})
    def {{ f.name }}(self) -> {{ f.fget_annotation }}:
        {{f.fget}}
    {% endfor %}

{% endfor %}






from typing import Any, Optional

from attr import define


@define
class BuiltinScalar:
    tp: type
    default_value: Any
    graphql_name: str


class _BuiltinScalars:
    INT = BuiltinScalar(int, 0, graphql_name="Int")
    FLOAT = BuiltinScalar(float, 0.0, graphql_name="Float")
    STRING = BuiltinScalar(str, " - ", graphql_name="String")
    ID = BuiltinScalar(str, "9b2a0828-880d-4023-9909-de067984523c", graphql_name="ID")
    BOOLEAN = BuiltinScalar(bool, False, graphql_name="Boolean")
    UUID = BuiltinScalar(str, "9b2a0828-880d-4023-9909-de067984523c", graphql_name="UUID")

    @classmethod
    def __iter__(cls):
        for name, member in cls.__dict__.items():
            if name.isupper():
                yield member

    def by_graphql_name(self, name: str) -> Optional[BuiltinScalar]:
        for scalar in self:
            if scalar.graphql_name == name:
                return scalar

    def keys(self) -> list[str]:
        return [scalar.graphql_name for scalar in self]


BuiltinScalars = _BuiltinScalars()



from __future__ import annotations

from typing import Any, ClassVar, Generic, Optional, TypeVar

from PySide6.QtCore import QObject, Signal
from PySide6.QtQuick import QQuickItem

from qtgql.codegen.py.runtime.environment import get_gql_env
from qtgql.gqltransport.client import GqlClientMessage
from qtgql.tools import qproperty, slot

T_QObject = TypeVar("T_QObject", bound=QObject)


class QSingletonMeta(type(QObject)):  # type: ignore
    def __init__(cls, name, bases, dict):
        super().__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kw)
        return cls.instance


class BaseQueryHandler(Generic[T_QObject], QObject, metaclass=QSingletonMeta):
    """Each handler will be exposed to QML and."""

    instance: "ClassVar[Optional[BaseQueryHandler]]" = None
    ENV_NAME: ClassVar[str]
    operationName: ClassVar[str]
    _message_template: ClassVar[GqlClientMessage]

    graphqlChanged = Signal()
    dataChanged = Signal()
    completedChanged = Signal()
    errorChanged = Signal()

    signalNames = ("graphqlChanged", "dataChanged", "completedChanged", "errorChanged")

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        name = self.__class__.__name__
        self.setObjectName(name)
        self._query: str = ""
        self._completed: bool = False
        self._data: Optional[T_QObject] = None
        self.environment = get_gql_env(self.ENV_NAME)
        self.environment.add_query_handler(self)
        self._consumers_count: int = 0
        self._operation_on_the_fly: bool = False

    def unconsume(self) -> None:
        self._consumers_count -= 1

    def consume(self) -> None:
        # if it is the first consumer fetch the data here.
        if not self._operation_on_the_fly and not self._completed:
            self.fetch()

        self._consumers_count += 1

    @property
    def message(self) -> GqlClientMessage:
        return self._message_template

    @qproperty(QObject, notify=dataChanged)
    def data(self) -> Optional[QObject]:
        return self._data

    @qproperty(bool, notify=completedChanged)
    def completed(self):
        return self._completed

    def fetch(self) -> None:
        self._operation_on_the_fly = True
        self.environment.client.execute(self)  # type: ignore

    def on_data(self, message: dict) -> None:  # pragma: no cover
        # real is on derived class.
        raise NotImplementedError

    def on_completed(self) -> None:
        self._completed = True
        self.completedChanged.emit()

    def on_error(self, message: list[dict[str, Any]]) -> None:  # pragma: no cover
        # This (unlike `on_data` is not implemented for real)
        raise NotImplementedError(message)


class UseQueryABC(QQuickItem):
    """Concrete implementation in the template."""

    ENV_NAME: ClassVar[str]

    # signals
    operationNameChanged = Signal()

    def __init__(self, parent: Optional[QQuickItem] = None):
        super().__init__(parent)
        self.destroyed.connect(self.unconsume)  # type: ignore
        self._operationName: Optional[str] = None
        self.env = get_gql_env(self.ENV_NAME)
        self.handler: Optional[BaseQueryHandler] = None

    @slot
    def set_operationName(self, graphql: str) -> None:
        self._operationName = graphql
        self.handler = self.env.get_handler(graphql)
        self.handler.consume()
        self.operationNameChanged.emit()  # type: ignore

    @qproperty(str, fset=set_operationName, notify=operationNameChanged)
    def operationName(self):
        return self._operationName

    @slot
    def unconsume(self) -> None:
        assert self.handler
        self.handler.unconsume()



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Callable, NewType, Optional, Type, TypedDict, TypeVar, Union

import graphql
from graphql import OperationType
from graphql.type import definition as gql_def

from qtgql.codegen.py.compiler.builtin_scalars import BuiltinScalars
from qtgql.codegen.py.compiler.query import QueryHandlerDefinition
from qtgql.codegen.py.compiler.template import (
    TemplateContext,
    handlers_template,
    schema_types_template,
)
from qtgql.codegen.py.objecttype import (
    EnumMap,
    EnumValue,
    GqlEnumDefinition,
    GqlFieldDefinition,
    GqlTypeDefinition,
    GqlTypeHinter,
)
from qtgql.codegen.utils import anti_forward_ref
from qtgql.exceptions import QtGqlException

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.compiler.config import QtGqlConfig

introspection_query = graphql.get_introspection_query(descriptions=True)


OperationName = NewType("OperationName", str)


class GeneratedNamespace(TypedDict):
    handlers: str
    objecttypes: str


class SchemaEvaluator:
    def __init__(self, config: QtGqlConfig):
        self.config = config
        self._generated_types: dict[str, GqlTypeDefinition] = {}
        self._generated_enums: EnumMap = {}
        self._fragments_store: dict[int, str] = {}
        self._query_handlers: dict[OperationName, QueryHandlerDefinition] = {}

    @cached_property
    def schema_definition(self) -> graphql.GraphQLSchema:
        with open(self.config.graphql_dir / "schema.graphql", "r") as f:
            return graphql.build_schema(f.read())

    @property
    def query_type(self) -> GqlTypeDefinition:
        query_type = self.schema_definition.query_type
        assert query_type
        return self._generated_types[query_type.name]

    def _evaluate_field_type(self, t: gql_def.GraphQLType) -> GqlTypeHinter:
        if non_null := is_non_null_definition(t):
            # There are no optionals in qtgql, we use default values.
            # By default, everything in graphql is optional,
            # so NON_NULL doesn't really make a difference,
            return self._evaluate_field_type(non_null.of_type)

        if list_def := is_list_definition(t):
            return GqlTypeHinter(type=list, of_type=(self._evaluate_field_type(list_def.of_type),))
        elif scalar_def := is_scalar_definition(t):
            if builtin_scalar := BuiltinScalars.by_graphql_name(scalar_def.name):
                return GqlTypeHinter(type=builtin_scalar)
            else:
                return GqlTypeHinter(type=self.config.custom_scalars[scalar_def.name])
        elif enum_def := is_enum_definition(t):
            return GqlTypeHinter(
                type=anti_forward_ref(name=enum_def.name, type_map=self._generated_enums)
            )
        elif obj_def := is_object_definition(t):
            return GqlTypeHinter(
                type=anti_forward_ref(name=obj_def.name, type_map=self._generated_types)
            )
        elif union_def := is_union_definition(t):
            return GqlTypeHinter(
                type=Union,
                of_type=tuple(
                    GqlTypeHinter(
                        type=anti_forward_ref(name=possible.name, type_map=self._generated_types)
                    )
                    for possible in self.schema_definition.get_possible_types(union_def)
                ),
            )

        raise NotImplementedError(f"type {t} not supported yet")  # pragma: no cover

    def _evaluate_field(self, name: str, field: gql_def.GraphQLField) -> GqlFieldDefinition:
        """we don't really know what is the field type just it's name."""
        return GqlFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._generated_types,
            scalars=self.config.custom_scalars,
            description=field.description,
            enums=self._generated_enums,
        )

    def _evaluate_object_type(
        self, type_: gql_def.GraphQLObjectType
    ) -> Optional[GqlTypeDefinition]:
        t_name: str = type_.name
        if t_name.startswith("__"):
            return None
        if evaluated := self._generated_types.get(t_name, None):
            return evaluated

        concrete = GqlTypeDefinition(
            name=t_name,
            docstring=type_.description,
            fields=[self._evaluate_field(name, field) for name, field in type_.fields.items()],
        )
        return concrete

    def _evaluate_enum(self, enum: gql_def.GraphQLEnumType) -> Optional[GqlEnumDefinition]:
        name: str = enum.name

        if self._generated_enums.get(name, None):
            return None

        return GqlEnumDefinition(
            name=name,
            members=[
                EnumValue(name=name, description=val.description or "")
                for name, val in enum.values.items()
            ],
        )

    def parse_schema_concretes(self) -> None:
        for name, type_ in self.schema_definition.type_map.items():
            if name.startswith("__"):
                continue

            if object_definition := is_object_definition(type_):
                if object_type := self._evaluate_object_type(object_definition):
                    self._generated_types[object_type.name] = object_type
            elif enum_def := is_enum_definition(type_):
                if enum := self._evaluate_enum(enum_def):
                    self._generated_enums[enum.name] = enum

    def parse_queries(self) -> None:
        with open(self.config.graphql_dir / "operations.graphql", "r") as f:
            queries = graphql.parse(f.read())

        if errors := graphql.validate(self.schema_definition, queries):
            raise QtGqlException([error.formatted for error in errors])

        for definition in queries.definitions:
            field_name = definition.selection_set.selections[0].name.value  # type: ignore
            field = None
            if definition.operation is OperationType.QUERY:  # type: ignore
                for field_impl in self.query_type.fields:
                    if field_impl.name == field_name:
                        field = field_impl
                assert field
                op_name = OperationName(definition.name.value)  # type: ignore
                self._query_handlers[op_name] = QueryHandlerDefinition(
                    query=graphql.print_ast(definition),
                    name=op_name,
                    field=field,
                    directives=definition.directives,  # type: ignore
                )

    def dumps(self) -> GeneratedNamespace:
        """:return: The generated schema module as a string."""
        self.parse_schema_concretes()
        self.parse_queries()
        context = TemplateContext(
            enums=list(self._generated_enums.values()),
            types=[
                t for name, t in self._generated_types.items() if name not in BuiltinScalars.keys()
            ],
            queries=list(self._query_handlers.values()),
            config=self.config,
        )
        return GeneratedNamespace(
            handlers=handlers_template(context), objecttypes=schema_types_template(context)
        )

    def dump(self):
        """:param file: Path to the directory the codegen would dump to."""
        for fname, content in self.dumps().items():
            with open(self.config.graphql_dir / (fname + ".py"), "w") as fh:
                fh.write(content)


T_Definition = TypeVar("T_Definition", bound=gql_def.GraphQLType)


def definition_identifier_factory(
    expected: Type[T_Definition],
) -> Callable[[gql_def.GraphQLType], Optional[T_Definition]]:
    def type_guarder(definition: gql_def.GraphQLType) -> Optional[T_Definition]:
        if isinstance(definition, expected):
            return definition

    return type_guarder


is_object_definition = definition_identifier_factory(gql_def.GraphQLObjectType)
is_enum_definition = definition_identifier_factory(gql_def.GraphQLEnumType)
is_list_definition = definition_identifier_factory(gql_def.GraphQLList)
is_scalar_definition = definition_identifier_factory(gql_def.GraphQLScalarType)
is_union_definition = definition_identifier_factory(gql_def.GraphQLUnionType)

is_non_null_definition = definition_identifier_factory(gql_def.GraphQLNonNull)



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from qtgql.gqltransport.client import HandlerProto

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.runtime.queryhandler import BaseQueryHandler


class NetworkLayerProto(Protocol):
    def execute(self, handler: HandlerProto) -> None:
        """accepts a handler and expected to call the handler's `on_data` /
        `on_error` / 'on_completed' when the operation is completed."""


class QtGqlEnvironment:
    """Encapsulates a schema interaction.

    The schema **must** be coherent with the schema use by the code
    generator. This class is used by the codegen and not to be
    instantiated directly.
    """

    def __init__(self, client: NetworkLayerProto, name: str):
        """
        :param client: The network layer for communicated the GraphQL server,
        all the generated handlers for this environment  would use this layer.
        :param name: This would be used to retrieve this environment from the env map by
        the generated handlers based on the configurations.
        """
        self.client = client
        self._query_handlers: dict[str, BaseQueryHandler] = {}
        self.name = name

    def add_query_handler(self, handler: BaseQueryHandler) -> None:
        self._query_handlers[handler.objectName()] = handler

    def get_handler(self, operationName: str) -> BaseQueryHandler:
        return self._query_handlers[operationName]


_ENV_MAP: dict[str, QtGqlEnvironment] = {}
"""In the future this would be usefully if you want to use different schemas,
maybe with a directive."""


def set_gql_env(env: QtGqlEnvironment) -> None:
    _ENV_MAP[env.name] = env


def get_gql_env(name: str) -> QtGqlEnvironment:
    return _ENV_MAP[name]



import re
from typing import Optional


def get_operation_name(query: str) -> Optional[str]:
    if match := re.search(r"(subscription|mutation|query)(.*?({|\())", query):
        return match.group(2).replace(" ", "").strip("{").strip("(")



from __future__ import annotations

import types
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Generic, Optional, Type, TypeVar, Union

import attrs
from PySide6.QtCore import QAbstractListModel, QByteArray, QEventLoop, QModelIndex, Qt, Signal
from typing_extensions import Self, dataclass_transform

from qtgql.exceptions import QtGqlException
from qtgql.tools import slot
from qtgql.utils.typingref import UNSET, TypeHinter

IS_GQL = "is_gql"
IS_ROLE = "is_role"
OLD_NAME = "old_name"

__all__ = ["role", "get_base_type", "GenericModel", "RoleDoesNotExist", "asdict"]


def asdict(t: _BaseType) -> dict:
    return attrs.asdict(t)  # type: ignore


def role(
    default=None,
    factory=UNSET,
):
    """role is optional by default."""
    if factory is not UNSET:
        default = UNSET
    return attrs.field(
        default=attrs.NOTHING if default is UNSET else default,
        factory=None if factory is UNSET else factory,
        metadata={IS_ROLE: True},
    )


def field_is(key: str, field):
    metadata = getattr(field, "metadata", {})
    return metadata.get(key, False)


T_V = TypeVar("T_V")


@dataclass_transform(
    field_descriptors=(role,),
    kw_only_default=True,
)
class BaseTypeMeta(type):
    def __new__(mcs, name, bases, ns):
        super_new = super().__new__

        # exclude BaseType itself
        parents = [b for b in bases if isinstance(b, BaseTypeMeta)]
        cls = super_new(mcs, name, bases, ns)
        if not parents:
            # remove annotations to avoid global reference errors.
            cls.__annotations__ = {}
            return cls
        if not attrs.has(cls):
            assert cls not in cls.__types_map__.values()
            cls: type[_BaseType] = attrs.define(cls, slots=True, kw_only=True)
            cls.__types_map__[cls.__name__] = cls
            roles = RoleMapper()
            for role_num, field in enumerate(attrs.fields(cls), int(Qt.ItemDataRole.UserRole)):  # type: ignore
                # assign role and check if not exists
                if field_is(IS_ROLE, field):
                    role_ = Role(
                        name=field.name, num=role_num, type_map=cls.__types_map__, type=field.type
                    )
                    # fill the role manager
                    roles.by_num[role_num] = role_
                    roles.by_name[role_.name] = role_
                    roles.qt_roles[role_.num] = role_.qt_name

            cls.Model = GenericModel.from_role_defined(cls)
            cls.Model.__roles__ = roles

        return cls


class _BaseType(metaclass=BaseTypeMeta):
    if TYPE_CHECKING:  # pragma: no cover
        __types_map__: dict[str, _BaseType]
        Model: ClassVar[type[GenericModel[Self]]]  # type: ignore

    def asdict(self) -> dict:
        return attrs.asdict(self)  # type: ignore


_TBaseType = TypeVar("_TBaseType", bound=_BaseType)

count = 1


def get_base_type() -> type[_BaseType]:
    """:return: BaseType to use for generating `GenericModel`'s"""
    global count
    cls: type[_BaseType] = types.new_class(
        name=f"BaseType{count}", kwds={"metaclass": BaseTypeMeta}
    )
    # we need to inject this here in order for `get_type_hints()` to work.
    cls.__types_map__ = {GenericModel.__name__: GenericModel}  # type: ignore
    count += 1
    return cls


class Role:
    """Provides metadata about a field in a define_roles decorated class."""

    __slots__ = ("num", "name", "qt_name", "_type", "str_type", "type_map")

    def __init__(self, num: int, name: str, type: Union[TypeHinter, str], type_map: dict):
        self.num = num
        self.name = name
        self.qt_name = QByteArray(name.encode("utf-8"))
        self.type_map = type_map
        if isinstance(type, str):
            self.str_type = type
        else:
            self._type = TypeHinter.from_annotations(type)

    @property
    def type(self) -> TypeHinter:
        try:
            return self._type
        except AttributeError:
            self._type = TypeHinter.from_string(self.str_type, ns=self.type_map)
            return self._type


class RoleMapper:
    """A container that maps the roles of a defined class each map has a
    certain usage in the future."""

    def __init__(self):
        # this is the real name of the field
        # how the class would be created
        self.by_name: dict[str, Role] = {}
        # number is  a qt role 256+
        # this is how qml will call the data() method
        self.by_num: dict[int, Role] = {}
        # just to return to qt method roleNames()
        self.qt_roles: dict[int, QByteArray] = {}

    @cached_property
    def children(self) -> dict[str, type[_BaseType]]:
        # mapping all the roledefined to know for what
        # to initialize a genericModel
        ret = {}
        for role_ in self.by_name.values():
            # lists must be child models
            if role_.type.type is GenericModel:
                child_type = role_.type.of_type[0].type
                ret[role_.name] = child_type
        return ret


class RoleDoesNotExist(QtGqlException):
    ...


class GenericModel(Generic[_TBaseType], QAbstractListModel):
    """Contains logic autogenerated models that extends `_BaseType` You don't
    need to use this class directly.

    Although you would need it to annotate "child" roles. i.e:

    ```python
    from qtgql.tools.itemsystem import get_base_type, GenericModel

    BaseType = get_base_type()
    class SomeModel(BaseType):
         other_model: GenericModel['OtherModel']
    ```
    """

    __roledefined_type__: type[_TBaseType]
    __roles__: RoleMapper
    layoutAboutToBeChanged: Signal
    layoutChanged: Signal
    dataChanged: Callable[[QModelIndex, QModelIndex, Optional[list[int]]], None]

    def __init__(self, *, data: Optional[Union[list[dict], dict]] = None, parent=None):
        super().__init__(parent)
        self.rowsAb = None
        self.parent_model: Optional[GenericModel] = parent
        self.type_ = self.__roledefined_type__
        self._data: list[_TBaseType] = []
        self._child_models: list[GenericModel[Any]] = []
        if data:
            self.initialize_data(data)

    def initialize_data(self, data: Union[list[dict], dict]) -> None:
        """:param data: data to generate the model"""
        self.layoutAboutToBeChanged.emit()
        QEventLoop().processEvents()
        # search for children and initialize them as models
        for node in data:
            if children := self.__roles__.children:
                for name, child in children.items():
                    child_data = node[name]
                    child_model = child.Model(data=child_data, parent=self)
                    node[name] = child_model
                    #  save child models for further use.
                    self._child_models.append(child_model)

        # initialize list of attrs classes
        for node in data:
            node = self.type_(**node)  # type: ignore
            self._data.append(node)

        self.layoutChanged.emit()

    def data(self, index, role=...):
        if index.row() < len(self._data) and index.isValid():
            try:
                return getattr(self._data[index.row()], self.__roles__.by_num[role].name, None)
            except KeyError as exc:
                if role in (
                    -1,
                    Qt.ItemDataRole.DisplayRole,
                    Qt.ItemDataRole.ToolTipRole,
                    Qt.ItemDataRole.StatusTipRole,
                    Qt.ItemDataRole.WhatsThisRole,
                    Qt.ItemDataRole.SizeHintRole,
                    Qt.ItemDataRole.FontRole,
                    Qt.ItemDataRole.BackgroundRole,
                    Qt.ItemDataRole.ForegroundRole,
                    Qt.ItemDataRole.DecorationRole,
                    Qt.ItemDataRole.TextAlignmentRole,
                    Qt.ItemDataRole.CheckStateRole,
                ):
                    return None
                # resolvers should be pre-evaluated when the model updated
                raise RoleDoesNotExist(
                    f"role {role} of type {self.type_} at index: [{index}] "
                    f"is not a valid role!\n"
                    f"options are: {self.__roles__.qt_roles}"
                ) from exc

        return None

    def setData(self, index, value, role=...):
        if index.row() < len(self._data) and index.isValid():
            setattr(self._data[index.row()], self.__roles__.by_num[role].name, value)
            return True
        return False

    def flags(self, index):
        if index.row() < len(self._data) and index.isValid():
            return Qt.ItemFlag.ItemIsEditable
        return Qt.ItemFlag.NoItemFlags

    def roleNames(self) -> dict:
        return self.__roles__.qt_roles

    def rowCount(self, parent=...) -> int:
        return len(self._data)

    def append(self, node: _TBaseType) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()

    @slot
    def clear(self) -> None:
        if self._data:
            self.beginRemoveRows(self.index(0).parent(), 0, self.rowCount())
            for child in self._child_models:
                child.clear()
            self._data = []
            self.endRemoveRows()

    @slot
    def get_by_index(self, row: int, key: str) -> Any:
        return self.data(self.index(row), self.__roles__.by_name[key].num)

    # CLASS METHODS
    @classmethod
    def from_role_defined(
        cls, type_: type[_BaseType], parent: Optional[Type[GenericModel]] = None
    ) -> type[GenericModel]:
        bases = (cls,)
        if parent:
            bases = (parent,)
        ret = type(cls.name_by_type(type_), bases, {"__roledefined_type__": type_})
        assert issubclass(ret, GenericModel)
        return ret

    @classmethod
    def name_by_type(cls, type_: type) -> str:
        return type_.__name__ + "Model"



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
    get_type_hints,
    overload,
)

from PySide6.QtCore import Property, Signal

from qtgql.utils.typingref import TypeHinter

__all__ = ["qproperty"]

R = TypeVar("R")
if TYPE_CHECKING:  # pragma: no cover

    class qproperty_(property, Generic[R]):
        fget: Callable[[Any], R]
        fset: Optional[Callable[[Any, R], None]]
        fdel: Optional[Callable[[Any], None]]

        def __new__(
            cls,
            fget: Callable[[Any], R],
            fset: Optional[Callable[[Any, R], None]] = ...,
            fdel: Optional[Callable[[Any], None]] = ...,
        ) -> qproperty_[R]:
            ...

        @overload  # type: ignore
        def __get__(self, obj: None, type_: Optional[type] = ...) -> qproperty_[R]:
            ...

        @overload
        def __get__(self, obj: object, type_: Optional[type] = ...) -> R:
            ...

        def __set__(self, obj: Any, value: R) -> None:
            ...


def qproperty(
    type: Any = None,
    constant: bool = False,
    fset: Optional[Callable] = None,
    notify: Optional[Signal] = None,
) -> Callable[[Callable[[Any], R]], qproperty_[R]]:
    type_ = type

    def wrapper(func: Callable[[Any], R]) -> qproperty_[R]:
        nonlocal type_
        if not type_:
            try:
                annotation = get_type_hints(func)["return"]
                ret = TypeHinter.from_annotations(annotation)
                if ret.is_optional():
                    type_ = ret.of_type[0].type
                else:
                    type_ = ret.type
            except NameError:
                type_ = "QVariant"
        return Property(type=type_, fset=fset, notify=notify, constant=constant)(func)  # type: ignore

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qtgql',
 'qtgql.codegen',
 'qtgql.codegen.py',
 'qtgql.codegen.py.compiler',
 'qtgql.codegen.py.runtime',
 'qtgql.codegen.py.templates',
 'qtgql.ext',
 'qtgql.gqltransport',
 'qtgql.tools',
 'qtgql.utils']

package_data = \
{'': ['*']}

install_requires = \
['PySide6>=6.4.2,<7.0.0', 'attrs>=22.2.0,<23.0.0']

extras_require = \
{'codegen': ['graphql-core>=3.2.3,<4.0.0',
             'jinja2>=3.1.2,<4.0.0',
             'typer[all]>=0.7.0,<0.8.0',
             'toml>=0.10.2,<0.11.0']}

entry_points = \
{'console_scripts': ['qtgql = qtgql.codegen.cli:entrypoint']}

setup_kwargs = {
    'name': 'qtgql',
    'version': '0.107.0',
    'description': 'Qt framework for building graphql driven QML applications',
    'long_description': '###  Qt framework for building graphql driven QML applications\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/qtgql?style=for-the-badge)](https://pypi.org/project/qtgql/)\n[![PyPI](https://img.shields.io/pypi/v/qtgql?style=for-the-badge)](https://pypi.org/project/qtgql/)\n[![GitHub Workflow Status (with branch)](https://img.shields.io/github/actions/workflow/status/nrbnlulu/qtgql/tests.yml?branch=main&style=for-the-badge)\n](https://github.com/nrbnlulu/qtgql/actions/workflows/tests.yml)\n[![Codecov](https://img.shields.io/codecov/c/github/nrbnlulu/qtgql?style=for-the-badge)](https://app.codecov.io/gh/nrbnlulu/qtgql)\n[![Discord](https://img.shields.io/discord/1067870318301032558?label=discord&style=for-the-badge)](https://discord.gg/5vmRRJp9fu)\n\n\n### Disclaimer\nThis project is currently under development, and **it is not** production ready,\nYou can play-around and tell us what is wrong / missing / awesome :smile:.\n\n[Visit the docs for more info](https://nrbnlulu.github.io/qtgql/)\n\n\n### Features\n#### Codegen (introspection compiler)\n- [x] object types, for each field there is a corresponding `Property`\n- [x] enums\n- [x] custom scalars\n- [x] query handlers\n#### Runtime\n- [x] "Qt-native" graphql-transport-ws network manager (supports subscriptions).\n#### Helpers\n- [x] generic models that get created from dictionaries (with update, pop, insert implemented by default)\n- [x] `Property` classes that are accessible from QML, with dataclasses  syntax (using attrs)\n- [x] `@slot` - decorator to be replaced with `QtCore.Slot()` that get types from type hints.\n\n### Future vision\n- Code generation from schema inspection\nIdeally every graphql type would be a `QObject` with `Property` for each field.\n- possibly generate C++ bindings from schema inspection\n- Query only what defined by the user (similar to how relay does this)\n- Auto mutations\n- Subscriptions\n\n### "Just build a web based UI"\nQt-QML IMO is a big game changer\n- you get native performance\n- UI code is very clear and declarative\n- Easy to customize\n- Easy to learn\n\nOne of the big disadvantages in Qt-QML is that Qt-C++ API is very repetitive and hard to maintain\nfor data-driven applications.\n\nalthough it is tempting to just use `relay` or other `JS` graphql lib\nthere is a point where you would suffer from performance issues (react-native).\n',
    'author': 'Nir',
    'author_email': '88795475+nrbnlulu@users.noreply.github.com',
    'maintainer': 'Nir.J Benlulu',
    'maintainer_email': 'nrbnlulu@gmail.com',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



import importlib
from pathlib import Path
from typing import Optional

import rich
import toml
import typer

from qtgql.codegen.py.compiler.config import QtGqlConfig

console = rich.console.Console()


# these lines are covered coverage failed to catch them.
def _get_pyproject(p: Path) -> Optional[Path]:
    pproject = p / "pyproject.toml"
    if pproject.exists():
        return pproject
    return None  # pragma: no cover


def _find_pyproject(p: Path) -> Optional[Path]:
    if pp := _get_pyproject(p):
        return pp
    for parent in p.parents:  # pragma: no cover
        return _find_pyproject(parent)


def _get_app_import_path(root: Path = Path.cwd()) -> str:
    pyproject = _find_pyproject(root)
    assert pyproject
    pp = toml.load(pyproject)
    try:
        return pp["tool"][TOOL_NAME][QTGQL_CONFIG_KEY]
    except KeyError as e:
        raise KeyError(
            f"Could not find 'tool.{TOOL_NAME}' in your pyproject.toml.",
            f"or you haven't defined '{QTGQL_CONFIG_KEY}'",
            "Make sure you have configured your project properly",
        ) from e


app = typer.Typer(pretty_exceptions_show_locals=False)

TOOL_NAME = "qtgql"
QTGQL_CONFIG_KEY = "config"


def _get_config() -> QtGqlConfig:
    console.print("[bold blue]Running system checks")
    mod, conf = _get_app_import_path().split(":")
    return getattr(importlib.import_module(mod), conf)


@app.command()
def gen():
    """Generates types based on your `QtGqlConfig` configuration object."""
    with console.status("Fetching schema and generating types...") as s:
        config = _get_config()
        s.update("[bold blue]Configuration file loaded")
        config.generate()
    console.print(
        f"[bold green]Types were generated to [link={config.generated_types_dir.resolve()}]file://{config.generated_types_dir.resolve()}[/link] successfully!"
    )


@app.command()
def hotreload():  # pragma: no cover
    raise NotImplementedError


def entrypoint():  # pragma: no cover
    app()



from typing import Optional, Union
from PySide6.QtCore import Signal, QObject
from PySide6.QtQml import QmlElement, QmlSingleton
from qtgql.codegen.py.runtime.queryhandler import BaseQueryHandler, UseQueryABC
from qtgql.gqltransport.client import  GqlClientMessage, QueryPayload
from objecttypes import * # noqa


QML_IMPORT_NAME = "generated.{{context.config.env_name}}"
QML_IMPORT_MAJOR_VERSION = 1






{% for query in context.queries %}
@QmlElement
@QmlSingleton
class {{query.name}}(BaseQueryHandler[{{query.field.annotation}}]):
    ENV_NAME = "{{context.config.env_name}}"
    _message_template = GqlClientMessage(payload=QueryPayload(query="""{{query.query}}""", operationName="{{query.name}}"))
    def on_data(self, message: dict) -> None:
        parent = self.parent()
        if {{query.field.name}} := message.get('{{query.field.name}}', None):
            self._data = {{query.field.deserializer}}
            self.dataChanged.emit()

{% endfor %}


def init() -> None:
    {% for query in context.queries %}
    {{query.name}}()
    {% endfor %}
@QmlElement
class UseQuery(UseQueryABC):
    ENV_NAME = "{{context.config.env_name}}"